/*-----------------------------------------------------------------------------
 *  "gcc-plugin.h" must be the FIRST file to be included 
 *-----------------------------------------------------------------------------*/
#include "gcc-plugin.h"
#include "config.h"
#include <stdio.h>
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "tm_p.h"
#include "diagnostic.h"
#include "tree-flow.h"
#include "tree-pass.h"
#include "toplev.h"

/*-----------------------------------------------------------------------------
 *  Each plugin MUST define this global int to assert compatibility with GPL; 
 *  else the compiler throws a fatal error 
 *-----------------------------------------------------------------------------*/
int plugin_is_GPL_compatible;  

/*function declarations*/
static unsigned int inter_gimple_manipulation (void);

/*-----------------------------------------------------------------------------
 *  Structure of the pass we want to insert, identical to a regular ipa pass
 *-----------------------------------------------------------------------------*/
struct simple_ipa_opt_pass pass_plugin1 = {
  {
    SIMPLE_IPA_PASS,
    "callers",                            /*  name */
    NULL,                                 /*  gate */
    inter_gimple_manipulation,            /*  execute */
    NULL,                                 /*  sub */
    NULL,                                 /*  next */
    0,                                    /*  static pass number */
    TV_INTEGRATION,                       /*  tv_id */
    0,                                    /*  properties required */
    0,                                    /*  properties provided */
    0,                                    /*  properties destroyed */
    0,                                    /*  todo_flags start */
    0                                     /*  todo_flags end */
  }
};


/*-----------------------------------------------------------------------------
 *  This structure provides the information about inserting the pass in the
 *  pass manager. 
 *-----------------------------------------------------------------------------*/
struct register_pass_info pass_info1 = {
  &(pass_plugin1.pass),            /* Address of new pass, here, the 'struct
                                     opt_pass' field of 'gimple_opt_pass'
                                     defined above */
  "pta",                          /* Name of the reference pass for hooking up
                                     the new pass.   ??? */
  0,                              /* Insert the pass at the specified instance
                                     number of the reference pass. Do it for
                                     every instance if it is 0. */
  PASS_POS_INSERT_BEFORE           /* how to insert the new pass: before,
                                     after, or replace. Here we are inserting
                                     a pass names 'plug' after the pass named
                                     'pta' */
};

/*-----------------------------------------------------------------------------
 *  plugin_init is the first function to be called after the plugin is loaded
 *-----------------------------------------------------------------------------*/
  int
plugin_init(struct plugin_name_args *plugin_info,
    struct plugin_gcc_version *version)
{

  /*-----------------------------------------------------------------------------
   * Plugins are activiated using this callback 
   *-----------------------------------------------------------------------------*/
  register_callback (
      plugin_info->base_name,     /* char *name: Plugin name, could be any
                                     name. plugin_info->base_name gives this
                                     filename */
      PLUGIN_PASS_MANAGER_SETUP,  /* int event: The event code. Here, setting
                                     up a new pass */
      NULL,                       /* The function that handles event */
      &pass_info1);                /* plugin specific data */

  return 0;
}

/* Execute function */
static unsigned int inter_gimple_manipulation (void)
{
        struct cgraph_node *node;
        struct cgraph_edge *edge;

        if (!dump_file)
                return 1;


        fprintf (dump_file, "Print callers of a function : \n");
        /* Iterating over each basic block of a function */
        for (node = cgraph_nodes; node; node = node->next)
        {
                /* Nodes without a body, and clone nodes are not interesting. */
                if (!gimple_has_body_p (node->decl) || node->clone_of)
                        continue;

                push_cfun (DECL_STRUCT_FUNCTION (node->decl));

                fprintf (dump_file, "\n Function : ");
                fprintf (dump_file,"%s\t",cgraph_node_name(node));

		/* Check if the node i.e. a function has callers */
                if(node->callers)
                {
			/* Print the first caller */
                        fprintf (dump_file, "\t Caller :");
                        fprintf (dump_file,"%s",cgraph_node_name(node->callers->caller));

			/* Extract the next caller */
                        edge = node->callers->next_caller;

			/* Iterate over all the remaining callers */
                        while(edge)
                        {
                                fprintf (dump_file, "\t Caller :");
                                fprintf (dump_file,"%s",cgraph_node_name(edge->caller));
                                edge = edge->next_caller;
                        }
                }

                /* restoring the context by popping cfun. */
                pop_cfun ();

        }

        return 0;
}
